import { TransactionEnvelope } from "@saberhq/solana-contrib";
import { getATAAddress, getOrCreateATA, TOKEN_PROGRAM_ID, NATIVE_MINT, ASSOCIATED_TOKEN_PROGRAM_ID, } from "@saberhq/token-utils";
import { Keypair, SystemProgram } from "@solana/web3.js";
import { Token } from "@solana/spl-token";
import { findClaimStatusKey, findDistributorKey } from "./pda";
import { toBytes32Array } from "./utils";
export class MerkleDistributorWrapper {
    constructor(sdk, key, distributorATA, data) {
        this.sdk = sdk;
        this.program = sdk.program;
        this.key = key;
        this.distributorATA = distributorATA;
        this.data = data;
    }
    static async load(sdk, key) {
        const data = await sdk.program.account.merkleDistributor.fetch(key);
        return new MerkleDistributorWrapper(sdk, key, await getATAAddress({ mint: data.mint, owner: key }), data);
    }
    static async createDistributor(args) {
        var _a;
        const { root, tokenMint } = args;
        const { sdk } = args;
        const { provider } = sdk;
        const baseKey = (_a = args.base) !== null && _a !== void 0 ? _a : Keypair.generate();
        const [distributor, bump] = await findDistributorKey(baseKey.publicKey);
        const ixs = [];
        ixs.push(sdk.program.instruction.newDistributor(bump, toBytes32Array(root), args.maxTotalClaim, args.maxNumNodes, {
            accounts: {
                base: baseKey.publicKey,
                distributor,
                mint: tokenMint,
                payer: provider.wallet.publicKey,
                systemProgram: SystemProgram.programId,
            },
        }));
        const { address, instruction } = await getOrCreateATA({
            provider,
            mint: tokenMint,
            owner: distributor,
        });
        if (instruction) {
            ixs.push(instruction);
        }
        return {
            base: baseKey.publicKey,
            bump,
            distributor,
            distributorATA: address,
            tx: new TransactionEnvelope(provider, ixs, [baseKey]),
        };
    }
    async claimIX(args, payer) {
        const { amount, claimant, index, proof } = args;
        const [claimStatus, bump] = await findClaimStatusKey(index, this.key);
        return this.program.instruction.claim(bump, index, amount, proof.map((p) => toBytes32Array(p)), {
            accounts: {
                distributor: this.key,
                claimStatus,
                from: this.distributorATA,
                to: await getATAAddress({ mint: this.data.mint, owner: claimant }),
                claimant,
                payer,
                systemProgram: SystemProgram.programId,
                tokenProgram: TOKEN_PROGRAM_ID,
            },
        });
    }
    async claim(args) {
        const { provider } = this.sdk;
        const tx = new TransactionEnvelope(provider, [
            await this.claimIX(args, provider.wallet.publicKey),
        ]);
        const { instruction } = await getOrCreateATA({
            provider,
            mint: this.data.mint,
            owner: args.claimant,
        });
        if (instruction) {
            tx.instructions.unshift(instruction);
        }
        // check data.mint is wrapped sol or not
        // if wrapped sol, add unwrap instruction
        if (this.data.mint.equals(NATIVE_MINT)) {
            const tokenAccount = await Token.getAssociatedTokenAddress(ASSOCIATED_TOKEN_PROGRAM_ID, TOKEN_PROGRAM_ID, NATIVE_MINT, provider.wallet.publicKey);
            const unwrapIx = Token.createCloseAccountInstruction(TOKEN_PROGRAM_ID, tokenAccount, provider.wallet.publicKey, provider.wallet.publicKey, []);
            tx.instructions.push(unwrapIx);
        }
        return tx;
    }
    async getClaimStatus(index) {
        const [key] = await findClaimStatusKey(index, this.key);
        return this.program.account.claimStatus.fetch(key);
    }
    async reload() {
        this.data = await this.program.account.merkleDistributor.fetch(this.key);
    }
}
//# sourceMappingURL=wrapper.js.map